package rs.etf.epos;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Security;
import java.security.Signature;
import java.security.SignatureException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;

import static rs.etf.epos.Utility.loadBytes;

/**
 * @author Nemanja Kojic
 *
 */
public class DigitalSignatureProvider implements IDigitalSignatureProvider {

	protected final static int KEYSIZE = 512;
	protected String algorithm = "RSA";
	protected String provider = "SUN";
	
	protected String getAlgorithm() {
		return algorithm;
	}
	
	public DigitalSignatureProvider(String alg) throws NoSuchAlgorithmException {
		//Signature.getInstance(alg);
		this.algorithm = alg;
	}
	
	public DigitalSignatureProvider(String algorithm, String provider) throws NoSuchAlgorithmException {
		Signature.getInstance(algorithm, Security.getProvider(provider)); // checks validity of the given algorithm
		
		this.algorithm = algorithm;
		this.provider = provider;
	}
	
	protected byte[] getBytes(InputStream input) throws IOException {
	    return loadBytes(input);
	}
	
	/**
	 * @Throws {@link RuntimeException} if can not generate keys, but it should never happen. 
	 */
	@Override
	public KeyPair generateKeys(String keyType) {
		try {
			// Adjust secure random numbers generator
			SecureRandom secrandomGen = new SecureRandom();
			
			// Provide KeyPairGenerator with the previously adjusted secure random generator.
			KeyPairGenerator keygen = KeyPairGenerator.getInstance(keyType);
			keygen.initialize(KEYSIZE, secrandomGen);

			// Generate key pairs.
			return keygen.generateKeyPair();
		} catch(Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	
	@Override
	public byte[] generateSignature(byte[] contentToSign, PrivateKey privkey) throws SignatureException, NoSuchAlgorithmException, InvalidKeyException {
	    if (contentToSign == null) 
	        throw new IllegalArgumentException("Content to sign is null.");

	    Signature signAlg = Signature.getInstance(getAlgorithm());
	    signAlg.initSign(privkey);
	    signAlg.update(contentToSign);
	    return signAlg.sign();

	}
	
	/**
	 * Utility method for content signing 
	 * 
	 * @param contentToSign
	 * @param signedContentOut
	 * @param privkey
	 * @throws IOException 
	 * @throws IOException
	 * @throws NoSuchAlgorithmException 
	 * @throws SignatureException 
	 * @throws InvalidKeyException 
	 * @throws InvalidKeyException
	 * @throws SignatureException
	 */
	@Override
    public void signContent(byte[] contentToSign, OutputStream signedContentOut, PrivateKey privkey) throws IOException, InvalidKeyException, SignatureException, NoSuchAlgorithmException {
        byte[] signature = generateSignature(contentToSign, privkey);
        ObjectOutputStream oos = new ObjectOutputStream(signedContentOut);
        oos.writeObject(contentToSign);
        oos.writeObject(signature);
        oos.close();        
    }
	
	@Override
	public void signContent(byte[] contentToSign, OutputStream signedContentOut, PrivateKey privkey, PublicKey pubkey) throws InvalidKeyException, SignatureException,
    NoSuchAlgorithmException, IOException {
	    byte[] signature = generateSignature(contentToSign, privkey);
	    ObjectOutputStream oos = new ObjectOutputStream(signedContentOut);
	    if (pubkey != null) {
	        oos.writeObject(pubkey.getEncoded());
	    }
        oos.writeObject(contentToSign);
        oos.writeObject(signature);
        oos.close();
	}
	
	@Override
	public void signContent(InputStream contentToSign, OutputStream signedContentOut, PrivateKey privkey) throws IOException, InvalidKeyException, SignatureException, NoSuchAlgorithmException{
		byte[] contentBytes = getBytes(contentToSign);
		
		signContent(contentBytes, signedContentOut, privkey);
	}
	
	@Override
	public boolean verifySignedContent(byte[] contentToVerify, PublicKey pubKey)
	        throws InvalidKeyException, SignatureException, NoSuchAlgorithmException, IOException,
	        ClassNotFoundException {
	    ObjectInputStream ois = null;
        try {
            ois = new ObjectInputStream(new ByteArrayInputStream(contentToVerify));
            
            // Extract original content and signature
            byte[] originalContent = (byte[])ois.readObject();
            byte[] signature = (byte[]) ois.readObject();
            
            // Create verificaton algorithm
            Signature verifyAlg = Signature.getInstance(getAlgorithm());
            verifyAlg.initVerify(pubKey);
            
            verifyAlg.update(originalContent);
            return verifyAlg.verify(signature);
        } 
        finally {
            if (ois != null) ois.close();
        }
	}
	
	/**
	 * Assumes that the public key is a part of the given signed content. 
	 * It is up to implementation how the public key is stored in the content.
	 * 
	 * This implementation assumes that public key is the first array of bytes in the content.
	 * @throws NoSuchProviderException 
	 * @throws InvalidKeySpecException 
	 */
	@Override
	public boolean verifySignedContent(byte[] contentToVerify)
	throws InvalidKeyException, SignatureException, NoSuchAlgorithmException, IOException, ClassNotFoundException, InvalidKeySpecException, NoSuchProviderException {
	    ObjectInputStream ois = null;
	    try {
	        ois = new ObjectInputStream(new ByteArrayInputStream(contentToVerify));
	        
	        // Extract original content and signature
	        byte[] pubkeyBytes = (byte[]) ois.readObject();
	        PublicKey pubkey = decode(pubkeyBytes);
	        
	        byte[] originalContent = (byte[])ois.readObject();
	        byte[] signature = (byte[]) ois.readObject();

	        // Create verificaton algorithm
	        Signature verifyAlg = Signature.getInstance(this.algorithm, this.provider);
	        verifyAlg.initVerify(pubkey);

	        verifyAlg.update(originalContent);
	        return verifyAlg.verify(signature);
	    } 
	    finally {
	        if (ois != null) ois.close();
	    }
	}
	
	protected PublicKey decode(byte[] encodedKey) throws InvalidKeySpecException, NoSuchAlgorithmException, NoSuchProviderException {
	    X509EncodedKeySpec keySpec = new X509EncodedKeySpec(encodedKey);
	    KeyFactory factory = KeyFactory.getInstance(this.algorithm, this.provider);
	    return factory.generatePublic(keySpec);
	}
	
	/**
	 * @throws NoSuchProviderException 
	 * @throws InvalidKeySpecException 
	 * @throws SecurityException if License verification failed.
	 */
	@Override
	public byte[] extractOriginalContent(byte[] signedContent) throws InvalidKeyException, SignatureException, IOException, NoSuchAlgorithmException, ClassNotFoundException, NoSuchProviderException, InvalidKeySpecException {
	    byte[] result = null;
	    if (verifySignedContent(signedContent)) {
	        ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(signedContent));
	        ois.readObject(); // remove bytes belonging to the public key
	        result = (byte[])ois.readObject();
	        ois.close();
	    }
	    else {
	        throw new SecurityException("Licence could not be verified!");
	    }
	    return result;
	}

	@Override
	public byte[] extractOriginalContent(byte[] signedContent, PublicKey pubkey) throws IOException, NoSuchAlgorithmException, ClassNotFoundException,
	InvalidKeyException, SignatureException, NoSuchProviderException, InvalidKeySpecException {
	    byte[] result = null;
        if (verifySignedContent(signedContent, pubkey)) {
            ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(signedContent));
            result = (byte[])ois.readObject();
            ois.close();
        }
        else {
            throw new SecurityException("Licence could not be verified!");
        }
        return result;
	}
	
	
	protected PublicKey obtainPublicKey (byte[] signedContent) throws IOException, ClassNotFoundException {
	    ObjectInputStream ois = null;
	    try {
	        ois = new ObjectInputStream(new ByteArrayInputStream(signedContent));
	        return (PublicKey) ois.readObject();
	    } 
	    finally {
	        if (ois != null) ois.close();
	    }
	}

	
}
