package rs.etf.epos;

import static rs.etf.epos.Utility.loadBytes;
import static rs.etf.epos.Utility.toHex;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.InvalidKeyException;
import java.security.KeyPair;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SignatureException;
import static rs.etf.epos.Utility.*;
import org.junit.Test;
import static org.junit.Assert.*;

public class DSTest {

	static final String KEY_TYPE = "RSA";
	static final String SIGN_ALGORITHM = "SHA1withRSA";
//	static final String KEY_TYPE = "DSA";
//	static final String SIGN_ALGORITHM = "DSA";
	
	static byte[] computeFingerprint(InputStream is, String algorithmName) throws NoSuchAlgorithmException, IOException {
		MessageDigest algorithm = MessageDigest.getInstance(algorithmName);
		byte[] bytes = loadBytes(is);
		algorithm.reset();
		algorithm.update(bytes);
		return algorithm.digest();
	}
	
	@Test
	public void testMakingFingerprint() throws NoSuchAlgorithmException, FileNotFoundException, IOException {
		File msg = new File("src/message.txt");
		byte[] fingerprint = computeFingerprint(new FileInputStream(msg), "SHA-1");
		System.out.println("SHA-1: "+toHex(fingerprint));
		fingerprint = computeFingerprint(new FileInputStream(msg), "MD5");
		System.out.println("MD5:   " +toHex(fingerprint));
	}
	
	@Test 
	public void testKeyGeneration() throws NoSuchAlgorithmException, FileNotFoundException, IOException {
		IDigitalSignatureProvider idsp = new DigitalSignatureProvider(SIGN_ALGORITHM);
		KeyPair keys = idsp.generateKeys(KEY_TYPE);
		
		PrivateKey privateKey = keys.getPrivate();
		assertNotNull(privateKey);
		
		PublicKey publicKey = keys.getPublic();
		assertNotNull(publicKey);
		storeObject(new File("public.key"), publicKey);
		storeObject(new File("private.key"), privateKey);
		System.out.println(publicKey.toString());
		System.out.println(privateKey.toString());
	}
	
	@Test
	public void testDigitalSignatureGeneration() throws IOException, ClassNotFoundException, NoSuchAlgorithmException, InvalidKeyException, SignatureException {
		PrivateKey privateKey = (PrivateKey) loadObjects(new File("private.key"), 1)[0];
		File message = new File("src/message.txt");
		IDigitalSignatureProvider idsp = new DigitalSignatureProvider(SIGN_ALGORITHM);
		byte[] signature = idsp.generateSignature(loadBytes(new FileInputStream(message)), privateKey);
		System.out.println(toHex(signature));
	}
	
	@Test
	public void testDigitalSignature() throws IOException, ClassNotFoundException, NoSuchAlgorithmException, InvalidKeyException, SignatureException {
		PrivateKey privateKey = (PrivateKey) loadObjects(new File("private.key"), 1)[0];
		File message = new File("src/message.txt");
		IDigitalSignatureProvider idsp = new DigitalSignatureProvider(SIGN_ALGORITHM);
		byte[] signature = idsp.generateSignature(loadBytes(new FileInputStream(message)), privateKey);
		
		// Generate digitaly signed message
		ObjectOutputStream oos = null;
		try {
			oos = new ObjectOutputStream(new FileOutputStream("digitaly-signed-msg.txt"));
			oos.writeObject(loadBytes(new FileInputStream(message)));
			oos.writeObject(signature);
		} 
		finally {
			oos.close();
		}
		System.out.println(toHex(signature));
	}
	
	@Test
	public void verifySignedMessage() throws NoSuchAlgorithmException, IOException, ClassNotFoundException, InvalidKeyException, SignatureException {
		IDigitalSignatureProvider idsp = new DigitalSignatureProvider(SIGN_ALGORITHM);
		PublicKey publicKey = (PublicKey) loadObjects(new File("public.key"), 1)[0];
		boolean verified = idsp.verifySignedContent(loadBytes(new FileInputStream("digitaly-signed-msg.txt")), publicKey);
		assertTrue(verified);
	}
	
	@Test
	public void testCorruptedMessage() throws NoSuchAlgorithmException, InvalidKeyException, SignatureException, FileNotFoundException, IOException, ClassNotFoundException {
		IDigitalSignatureProvider idsp = new DigitalSignatureProvider(SIGN_ALGORITHM);
		PublicKey publicKey = (PublicKey) loadObjects(new File("public.key"), 1)[0];
		byte[] signedMsgBytes = loadBytes(new FileInputStream("digitaly-signed-msg.txt"));
		byte b = signedMsgBytes[100];
		
		signedMsgBytes[100] = 0;
		boolean verified = idsp.verifySignedContent(signedMsgBytes, publicKey);
		assertFalse(verified);
		
		signedMsgBytes[100] = b;
		verified = idsp.verifySignedContent(signedMsgBytes, publicKey);
		assertTrue(verified);
	}
	
	
}
